/*
 * detectMarkers.cpp
 *
 *  Created on: 12 Feb 2017
 *      Author: db12166
 */


/*
By downloading, copying, installing or using the software you agree to this
license. If you do not agree to this license, do not download, install,
copy or use the software.
                          License Agreement
               For Open Source Computer Vision Library
                       (3-clause BSD License)
Copyright (C) 2013, OpenCV Foundation, all rights reserved.
Third party copyrights are property of their respective owners.
Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
 * Neither the names of the copyright holders nor the names of the contributors
    may be used to endorse or promote products derived from this software
    without specific prior written permission.
This software is provided by the copyright holders and contributors "as is" and
any express or implied warranties, including, but not limited to, the implied
warranties of merchantability and fitness for a particular purpose are
disclaimed. In no event shall copyright holders or contributors be liable for
any direct, indirect, incidental, special, exemplary, or consequential damages
(including, but not limited to, procurement of substitute goods or services;
loss of use, data, or profits; or business interruption) however caused
and on any theory of liability, whether in contract, strict liability,
or tort (including negligence or otherwise) arising in any way out of
the use of this software, even if advised of the possibility of such damage.
 */


#include <opencv2/highgui.hpp>
#include <opencv2/aruco.hpp>
#include <opencv2/imgproc.hpp>
#include <opencv2/calib3d.hpp>

#include <iostream>
#include<vector>
#include <sstream>
#include <iomanip>
#include <fstream>
#include <algorithm>
#include <dirent.h>
#include <direct.h>


#define SSTR( x ) static_cast< std::ostringstream & >( \
		( std::ostringstream() << std::dec << x ) ).str()

using namespace std;
using namespace cv;

namespace {
const char* about = "Basic marker detection";
const char* keys  =
		"{p        |       | directory to scan videos}";
//const char* keys  =
//		"{d        |       | dictionary: DICT_4X4_50=0, DICT_4X4_100=1, DICT_4X4_250=2,"
//		"DICT_4X4_1000=3, DICT_5X5_50=4, DICT_5X5_100=5, DICT_5X5_250=6, DICT_5X5_1000=7, "
//		"DICT_6X6_50=8, DICT_6X6_100=9, DICT_6X6_250=10, DICT_6X6_1000=11, DICT_7X7_50=12,"
//		"DICT_7X7_100=13, DICT_7X7_250=14, DICT_7X7_1000=15, DICT_ARUCO_ORIGINAL = 16}"
//		"{v        |       | Input from video file, if ommited, input comes from camera }"
//		"{ci       | 0     | Camera id if input doesnt come from video (-v) }"
//		"{c        |       | Camera intrinsic parameters. Needed for camera pose }"
//		"{l        | 0.1   | Marker side lenght (in meters). Needed for correct scale in camera pose }"
//		"{dp       |       | File of marker detector parameters }"
//		"{r        |       | show rejected candidates too }";
}
//To determine files in dir.
bool has_suffix(const string& s, const string& suffix, const string& checkStr)
{
    return (s.size() >= suffix.size()) && equal(suffix.rbegin(), suffix.rend(), s.rbegin()) && (s.find(checkStr) != std::string::npos);
}
//List all files of a specific type in a directory.
//dirname = directory to search
//extStr =  ext of file
//strCmp character to search for in filename
//fileList - list of filenames (only)
//fullFileList - Path to files.
void listFilesInDir(string dirname,string extStr,string strCmp, vector<string>& fileList, vector<string>& fullFileList)
{


	DIR *dir = opendir(dirname.c_str());
	dirent *entry;

	if(!dir)
	{
		fileList.push_back("ERROR");
		fullFileList.push_back("ERROR");
		return;
	}



	//Loop through files.
	while((entry = readdir(dir))!=NULL)
	{
		if(has_suffix(entry->d_name, extStr, strCmp))
		{
			//cout << dirname +"\\"+ entry->d_name << endl;
			fileList.push_back(entry->d_name);
			fullFileList.push_back(dirname +"\\"+ entry->d_name);
		}
	}

	return;
}
/*Euc Distance*/
double distanceEuc(double x1, double y1,double z1, double x2, double y2,double z2)
{
	double x = x1 - x2;
	double y = y1 - y2;
	double z = z1 - z2;
	double dist;
	dist = pow(x, 2) + pow(y, 2) + pow(z, 2);
	dist = sqrt(dist);
	return dist;
}

/*Determine the camera coordinates from the rotation and translation vector.*/
cv::Point3f getCameraRelCoords(cv::Mat Rvec,cv::Mat Tvec)
{
	cv::Mat R;
	cv::Rodrigues(Rvec, R);
	cv::Mat cameraPose = -R.t() * (cv::Mat)Tvec;

	double x = cameraPose.at<double>(0, 0);
	double y = cameraPose.at<double>(0, 1);
	double z = cameraPose.at<double>(0, 2);

	return  cv::Point3f(x/1000,y/1000,z/1000);

}
// Checks if a matrix is a valid rotation matrix.
bool isRotationMatrix(cv::Mat R)
{
    Mat Rt;
    transpose(R, Rt);
    Mat shouldBeIdentity = Rt * R;
    Mat I = Mat::eye(3,3, shouldBeIdentity.type());

    return  norm(I, shouldBeIdentity) < 1e-6;

}

// Calculates rotation matrix to euler angles - based on rotm2Euler
cv::Vec3f rotationMatrixToEulerAngles(cv::Mat Rvec)
{

	cv::Mat R;
	cv::Rodrigues(Rvec, R);
    assert(isRotationMatrix(R));

    float sy = sqrt(R.at<double>(0,0) * R.at<double>(0,0) +  R.at<double>(1,0) * R.at<double>(1,0) );

    bool singular = sy < 1e-6; // If

    float x, y, z;
    if (!singular)
    {
        x = atan2(R.at<double>(2,1) , R.at<double>(2,2));
        y = atan2(-R.at<double>(2,0), sy);
        z = atan2(R.at<double>(1,0), R.at<double>(0,0));
    }
    else
    {
        x = atan2(-R.at<double>(1,2), R.at<double>(1,1));
        y = atan2(-R.at<double>(2,0), sy);
        z = 0;
    }
    return Vec3f(x, y, z);



}

double rad2Deg(double radVal)
{
	double degrees = radVal * 180 / CV_PI;
	return(degrees);
}

double deg2Rad(double degVal)
{
	double rads = (degVal/ 180) * CV_PI;
	return(rads);
}
/**
 */
static bool readCameraParameters(string filename, Mat &camMatrix, Mat &distCoeffs) {
	FileStorage fs(filename, FileStorage::READ);
	if(!fs.isOpened())
		return false;
	fs["camera_matrix"] >> camMatrix;
	fs["distortion_coefficients"] >> distCoeffs;
	return true;
}



/**
 */
static bool readDetectorParameters(string filename, Ptr<aruco::DetectorParameters> &params) {
	FileStorage fs(filename, FileStorage::READ);
	if(!fs.isOpened())
		return false;
	fs["adaptiveThreshWinSizeMin"] >> params->adaptiveThreshWinSizeMin;
	fs["adaptiveThreshWinSizeMax"] >> params->adaptiveThreshWinSizeMax;
	fs["adaptiveThreshWinSizeStep"] >> params->adaptiveThreshWinSizeStep;
	fs["adaptiveThreshConstant"] >> params->adaptiveThreshConstant;
	fs["minMarkerPerimeterRate"] >> params->minMarkerPerimeterRate;
	fs["maxMarkerPerimeterRate"] >> params->maxMarkerPerimeterRate;
	fs["polygonalApproxAccuracyRate"] >> params->polygonalApproxAccuracyRate;
	fs["minCornerDistanceRate"] >> params->minCornerDistanceRate;
	fs["minDistanceToBorder"] >> params->minDistanceToBorder;
	fs["minMarkerDistanceRate"] >> params->minMarkerDistanceRate;
	fs["doCornerRefinement"] >> params->doCornerRefinement;
	fs["cornerRefinementWinSize"] >> params->cornerRefinementWinSize;
	fs["cornerRefinementMaxIterations"] >> params->cornerRefinementMaxIterations;
	fs["cornerRefinementMinAccuracy"] >> params->cornerRefinementMinAccuracy;
	fs["markerBorderBits"] >> params->markerBorderBits;
	fs["perspectiveRemovePixelPerCell"] >> params->perspectiveRemovePixelPerCell;
	fs["perspectiveRemoveIgnoredMarginPerCell"] >> params->perspectiveRemoveIgnoredMarginPerCell;
	fs["maxErroneousBitsInBorderRate"] >> params->maxErroneousBitsInBorderRate;
	fs["minOtsuStdDev"] >> params->minOtsuStdDev;
	fs["errorCorrectionRate"] >> params->errorCorrectionRate;
	return true;
}



/**
 */
int main(int argc, char *argv[]) {
	CommandLineParser parser(argc, argv, keys);
	parser.about(about);

	int relDisThresh = 5;				// Dallan - Fisheye comp. To select Arruco tags closest to the camera.
	int dictionaryId ;
	bool showRejected ;
	bool estimatePose;
	float markerLength;
	int camId = 0;

	string dirSearch = parser.get<string>("p");
	cout<<dirSearch;

    //string dirSearch = "F:\\SphereVideo\\SpherePlusEv_SPHEREHOUSE_02_08_2017\\181YDPHW\\";
	string parentDir = dirSearch.substr(0, dirSearch.find_last_of("/\\"));
	string dirCal = parentDir + "\\calFiles";
	String annDir = dirSearch + "\\Annotated_Vids_AllDis";



	vector<string> fileList;
	vector<string> fullFileList;
	vector<string> fileWriteList;



	parser.printMessage();
	dictionaryId = 1;
	showRejected = false;
	estimatePose = true;
	markerLength = 0.126;//580;
//
//	if(argc < 2) {
//
//	}
//	else{
//
//		dictionaryId = parser.get<int>("d");
//		showRejected = parser.has("r");
//		estimatePose = parser.has("c");
//		markerLength = parser.get<float>("l");
//	}


	listFilesInDir(dirSearch,".mp4","S", fileList,fullFileList);
	listFilesInDir(dirSearch,".MP4","S", fileList,fullFileList);

	//fileList.push_back("F:\\SphereVideo\\Dallan_House_12_03_2017\\S1560002.mp4");
	//fileList.push_back("F:\\SphereVideo\\Dallan_House_12_03_2017\\S1560003.mp4");
	//fileList.push_back("F:\\SphereVideo\\Dallan_House_12_03_2017\\S1560004.mp4");
	//fileList.push_back("F:\\SphereVideo\\Michal_House_26_02_2017\\S1010006.mp4");
	//fileList.push_back("F:\\SphereVideo\\Michal_House_26_02_2017\\S1010003.mp4");
	int numFiles = fileList.size();
	//numFiles = 1;

	mkdir(annDir.c_str());

	for(int fI = 0; fI<numFiles;fI++)
	{
		String videoShort = fileList.at(fI);
		String video = fullFileList.at(fI);


		String saveVideo = annDir + "\\" + videoShort.substr(0, videoShort.length()-4) + "_Processed.avi";
		String saveFileAnn = annDir + "\\" + videoShort.substr(0, videoShort.length()-4) + "_Processed_Annotations.csv";
		cout<<videoShort<<endl;
		Ptr<aruco::DetectorParameters> detectorParams = aruco::DetectorParameters::create();
		char basePath[255] = "";
		_fullpath(basePath, argv[0], sizeof(basePath));
		string exePath = string(basePath);
		string parentDir = exePath.substr(0, exePath.find_last_of("/\\"));
		cout<<parentDir + "\\detector_params\\detectorParams.yml"<<endl;
		bool readOk = readDetectorParameters(parentDir + "\\detector_params\\detectorParams.yml", detectorParams);
		//	if(parser.has("dp")) {
		//		bool readOk = readDetectorParameters(parser.get<string>("dp"), detectorParams);
		if(!readOk) {
			cerr << "Invalid detector parameters file" << endl;
			return 0;
		}
		//	}
		detectorParams->doCornerRefinement = true; // do corner refinement in markers





		Ptr<aruco::Dictionary> dictionary =
				aruco::getPredefinedDictionary(aruco::PREDEFINED_DICTIONARY_NAME(dictionaryId));

		Mat camMatrix, distCoeffs;
		if(estimatePose) {
			bool readOk = readCameraParameters(dirCal + "\\calFile.yml", camMatrix, distCoeffs);
			if(!readOk) {
				cerr << "Cal File Error - Invalid camera file" << " "<<(dirCal + "\\calFile.yml")<< endl;
				return 0;
			}
		}

		VideoCapture inputVideo;

		if(!video.empty()) {
			inputVideo.open(video);
		} else {
			inputVideo.open(camId);
		}

		int ex = static_cast<int>(inputVideo.get(CAP_PROP_FOURCC));

		Size S = Size((int) inputVideo.get(CAP_PROP_FRAME_WIDTH),    // Acquire input size
				(int) inputVideo.get(CAP_PROP_FRAME_HEIGHT));

		Size op_size = Size((int) (S.width/2), (int)(S.height/2));
		int frame_num_total = inputVideo.get(CV_CAP_PROP_FRAME_COUNT);
		double totalTime = 0;
		int totalIterations = 0;


		int cnt = 0;
		VideoWriter oVideoWriter;
		ofstream csvAnnFile(saveFileAnn);
		int lastId=0;
		csvAnnFile<<SSTR("Vid_Time, TagID1, PosX1, PosY1, PosZ1, eulRot_X1, eulRot_Y1, eulRot_Z1, Euc2D_1,Euc3D_1,"
				" TagID2, PosX2, PosY2, PosZ2, eulRot_X2, eulRot_Y2, eulRot_Z2, Euc2D_2,Euc3D_2,"
				" TagID3, PosX3, PosY3, PosZ3, eulRot_X3, eulRot_Y3, eulRot_Z3, Euc2D_3,Euc3D_3,"
				" TagID4, PosX4, PosY4, PosZ4, eulRot_X4, eulRot_Y4, eulRot_Z4, Euc2D_4,Euc3D_4,"
				" TagID5, PosX5, PosY5, PosZ5, eulRot_X5, eulRot_Y5, eulRot_Z5,, Euc2D_5,Euc3D_5\n");

		int last_perc=-1;
		while(inputVideo.grab()) {

			Mat image, imageCopy;
			inputVideo.retrieve(image);

			double vidTime = inputVideo.get(CV_CAP_PROP_POS_MSEC);

			std::stringstream vidTime_ss;
			vidTime_ss <<vidTime/1e3;
			vidTime_ss.precision(2);

			string vidTimeStr =  "T="  + vidTime_ss.str();

			double tick = (double)getTickCount();

			int curr_frame = inputVideo.get(CV_CAP_PROP_POS_FRAMES);
			if(cnt==0)
			{
				oVideoWriter.open(saveVideo, ex,int(inputVideo.get(CAP_PROP_FPS)), op_size, true);
			}

			cnt++;

			vector< int > ids;
			vector< vector< Point2f > > corners, rejected;
			vector< Vec3d > rvecs, tvecs;

			// detect markers and estimate pose
			aruco::detectMarkers(image, dictionary, corners, ids, detectorParams, rejected);
			if(estimatePose && ids.size() > 0)
				aruco::estimatePoseSingleMarkers(corners, markerLength, camMatrix, distCoeffs, rvecs,
						tvecs);



			// draw results
			image.copyTo(imageCopy);
			if(ids.size() > 0) {
				aruco::drawDetectedMarkers(imageCopy, corners, ids);

				if(estimatePose) {
					String xStrStart= "Tag ID=";
					String xStr  = "";
					vector<cv::Point3f> relTagCoords;
					vector<cv::Vec3f> eulerRotAngles;

					for(unsigned int i = 0; i < ids.size(); i++)
					{




						aruco::drawAxis(imageCopy, camMatrix, distCoeffs, rvecs[i], tvecs[i],
								markerLength * 0.5f);

						//Get the euler angles.
						cv::Vec3f eulAngles = rotationMatrixToEulerAngles((cv::Mat)rvecs[i]);
						eulerRotAngles.push_back(eulAngles);
						//cout<<rad2Deg(eulAngles[0])<<", "<<rad2Deg(eulAngles[1])<<", "<<rad2Deg(eulAngles[2])<<endl;
						//double xyPose = rad2Deg(eulAngles[2]);	//Rotation around the z axis = xy plane angle. 0 degrees = Tag North
															//90* = Tag West, 180* = Tag South, 270* = Tag East.

						//cv::Point3f posOut = getCameraRelCoords( (cv::Mat)rvecs[i],  (cv::Mat)tvecs[i]);
						cv::Point3f posOut;
						posOut.x = tvecs[i][0];
						posOut.y = tvecs[i][1];
						posOut.z = tvecs[i][2];
						float xP = posOut.x;
						float yP = posOut.y;
						float zP = posOut.z;

						relTagCoords.push_back(posOut);
						if(i==0&&ids.size()<2)
						{
							if(abs(xP)<relDisThresh&&abs(yP)<relDisThresh)
							{
								xStr = SSTR("Only " <<setprecision(2)<< ids.at(i)<<" ["<<xP<<","<<yP<<"]");
								xStr += SSTR(" D = "<<setprecision(2)<<distanceEuc(xP, yP, 0,0, 0, 0));
								csvAnnFile<<vidTime<<", "<<ids.at(i)<<","<<xP<<", "<<yP<<", "<<zP<<", "
										<<eulAngles[0]<<", "<<eulAngles[1]<<", "<<eulAngles[2]<<", "
										<<distanceEuc(xP, yP, 0,0, 0, 0)<<", "<<distanceEuc(xP, yP, zP,0, 0, 0)<<"\n";
								lastId = ids[0];
							}

						}
						else
						{
							if(i==ids.size()-1)
							{
								//cout<<ids.size()<<" markers"<<endl;
								//Get the number of previous coordinates.
								vector<float> dis;
								vector<float> dis3D;

								float minDis = 0;
								int minDisI = 0;

								csvAnnFile<<vidTime;
								for(unsigned int pI = 0; pI<=i; pI++)
								{
									cv::Point3f prevCoord;
									prevCoord = relTagCoords[pI];


									dis.push_back(distanceEuc(prevCoord.x, prevCoord.y,0,0, 0, 0));
									dis3D.push_back(distanceEuc(prevCoord.x, prevCoord.y, prevCoord.z,0, 0, 0));


								}
								//cout<<"i = "<<i<<"  Min I - "<<minDisI<<endl;
								if(abs(relTagCoords[minDisI].x)<relDisThresh&&abs(relTagCoords[minDisI].y)<relDisThresh)
									xStr = SSTR( "Near "<<setprecision(2)<<ids.at(minDisI)<<" ["<<relTagCoords[minDisI].x<<","<<relTagCoords[minDisI].y<<"]");
									xStr += SSTR(" D = "<<setprecision(2)<<dis[minDisI]);

								//dis Sort ascending - record the indices
								vector<int> indexMap;
								for(unsigned int nI=0;nI<=i;nI++)
								{	indexMap.push_back(nI);}
								std::sort( indexMap.begin(),indexMap.end(), [&](int x,int y){return dis[x]<dis[y];} );

								//Record the last detected id
								lastId = ids[indexMap[0]];
								//Write to file.s
								for(unsigned int sI = 0; sI<=i; sI++)
								{
									cv::Vec3f rotAngles = eulerRotAngles[indexMap[sI]];
									//cout<<rotAngles[2]<<",";
									csvAnnFile<<", "<<ids.at(indexMap[sI])<<", "<<relTagCoords[indexMap[sI]].x<<", "<<relTagCoords[indexMap[sI]].y<<
											", "<<relTagCoords[indexMap[sI]].z<<", "
											<<rotAngles[0]<<", "<<rotAngles[1]<<", "<<rotAngles[2]
											<<", "<<dis[indexMap[sI]]<<", "<<dis3D[indexMap[sI]];



								}

								csvAnnFile<<"\n";

								//xStr = xStr + SSTR( ", "<<setprecision(2)<<ids.at(i)<<" ["<<xP<<","<<yP<<"]");
							}
						}


						putText(imageCopy, vidTimeStr + xStrStart + xStr,
								Point(10, 50), FONT_HERSHEY_SIMPLEX,1, Scalar(255, 0, 0), 2);

					}


					relTagCoords.empty();
				}

				double currentTime = ((double)getTickCount() - tick) / getTickFrequency();
				totalTime += currentTime;
				totalIterations++;

				int percentComplete = int(100*(float(curr_frame)/float(frame_num_total)));
				if((last_perc!=percentComplete) && (percentComplete>0) && (percentComplete % 10 == 0)) {
					//cout<< last_perc<<" "<<percentComplete<<endl;
					cout<< videoShort<<" Time elapsed: "<<int(vidTime/1000)<< "s ";
					cout <<" -- Detection Time = " << currentTime * 1000 << " ms "
							<< " % complete = "<<percentComplete <<"%"<< endl;
				}

				last_perc = percentComplete;
			}
			else{

				//Just record the last detected id.
				if(lastId>0){
					//cout<<"Last ID"<<lastId<<endl;
					csvAnnFile<<vidTime<<", "<<lastId<<","<<-500<<", "<<-500<<", "<<-500<<", "
							<<-500<<", "<<-500<<", "<<-500
							<<", "<<-1<<", "<<-1<<"\n";
					putText(imageCopy,SSTR("Last ID"<<lastId),
													Point(10, 50), FONT_HERSHEY_SIMPLEX,1, Scalar(255, 0, 0), 2);
				}
			}
			if(showRejected && rejected.size() > 0)
			{
				aruco::drawDetectedMarkers(imageCopy, rejected, noArray(), Scalar(100, 0, 255));

			}
			//waitKey(1);
			//if(key == 27) break;
			//imshow("out", imageCopy);

			Mat imageResize;
//			cv::Size img_size = imageCopy.size();
//			int img_width = img_size.width;
//			int img_height = img_size.height;
//			cv::Size new_size = cv::Size(img_width/2, img_height/2);
			resize(imageCopy,imageResize,op_size,INTER_CUBIC);
			oVideoWriter <<imageResize; //writer the frame into the file
			//char key = (char)waitKey(waitTime);
			//if(key == 27) break;
			//imshow("out", imageCopy);
			//		char key = (char)waitKey(waitTime);
			//		if(key == 27) break;



		}
		csvAnnFile.close();

	}

	return 0;
}

